//------------------------------------------------------------------------
//
// Author              : Jan Horn
// Email               : jhorn@global.co.za
// Website             : http://home.global.co.za/~jhorn
// Date                : 8 July 2001
// Version             : 1.0
// TMT Pascal version  : Vadim Bodrov (http://www.tmt.com)
// Description         : Uses the stencil buffer to clip drawing area of images
//
//------------------------------------------------------------------------
program Stencil3;

uses
  Windows,
  Messages,
  Strings,
  OpenGL,
  BMP;

const
  WND_TITLE = 'Stencil Buffer App by Jan Horn';
  FPS_TIMER = 1;                     // Timer to calculate FPS
  FPS_INTERVAL = 1000;               // Calculate FPS every 1000 ms

  SCALING_FACTOR1 = 16;
  SCALING_FACTOR2 = 64;

var
  h_Wnd  : HWND;                     // Global window handle
  h_DC   : HDC;                      // Global device context
  h_RC   : HGLRC;                    // OpenGL rendering context
  keys : Array[0..255] of Boolean;   // Holds keystrokes
  FPSCount : Integer = 0;            // Counter for FPS
  ElapsedTime : Integer;             // Elapsed time between frames

  // Textures
  BackgroundTex : glUint;

  // User vaiables
  KnotDL : glUint;

procedure glBindTexture(target: GLenum; texture: GLuint); stdcall; external 'opengl32.dll';


{------------------------------------------------------------------}
{  Function to create and draw a knot                              }
{------------------------------------------------------------------}
procedure createKnot(RADIUS1, RADIUS2, RADIUS3 : glFloat);
type glCoord = Record
       X, Y, Z : glFloat;
     end;
     glColors = Record
       R, G, B : glFloat;
     end;
     TmyObject = Record
       numfaces, numverts, numsides : Integer;
       faces    : Array [0..SCALING_FACTOR1 * SCALING_FACTOR2 * 4 * 4 - 1] of gluint;    //numfaces*numsides
       vertices : Array [0..SCALING_FACTOR1 * SCALING_FACTOR2 - 1] of glCoord;           //numverts
       Colors   : Array [0..SCALING_FACTOR1 * SCALING_FACTOR2 - 1] of GLColors;          //numverts
     end;
var Knot : TmyObject;
    I, J, Count1, Count2 : Integer;
    Alpha, Beta, Distance, MinDistance, Rotation : glFloat;
    X, Y, Z, dx, dy, dz : glFLoat;
    Value, modulus, dist : glFloat;
    index1, index2 : Integer;
begin
  Knot.numSides :=4;

  Alpha :=0;
  for Count2 :=0 to SCALING_FACTOR2-1 do
  begin
    Alpha :=Alpha + 2*PI / SCALING_FACTOR2;
    X := RADIUS2 * cos(2 * Alpha) + RADIUS1 * sin(Alpha);
    Y := RADIUS2 * sin(2 * Alpha) + RADIUS1 * cos(Alpha);
    Z := RADIUS2 * cos(3 * Alpha);
    dx := -2 * RADIUS2 * sin(2 * Alpha) + RADIUS1 * cos(Alpha);
    dy := 2 * RADIUS2 * cos(2 * Alpha) - RADIUS1 * sin(Alpha);
    dz := -3 * RADIUS2 * sin(3 * Alpha);
    Value := sqrt(dx * dx + dz * dz);
    Modulus := sqrt(dx * dx + dy * dy + dz * dz);

    Beta :=0;
    For Count1 :=0 to SCALING_FACTOR1-1 do
    begin
      Beta :=Beta + 2 * PI / SCALING_FACTOR1;

      Knot.vertices[Knot.numverts].x := X - RADIUS3 * (cos(Beta) * dz - sin(Beta) * dx * dy / Modulus) / Value;
      Knot.vertices[Knot.numverts].y := Y - RADIUS3 * sin(Beta) * Value / Modulus;
      Knot.vertices[Knot.numverts].z := Z + RADIUS3 * (cos(Beta) * dx + sin(Beta) * dy * dz / Modulus) / Value;


      dist := sqrt(Knot.vertices[Knot.numverts].x*Knot.vertices[Knot.numverts].x +
                   Knot.vertices[Knot.numverts].y*Knot.vertices[Knot.numverts].y +
                   Knot.vertices[Knot.numverts].z*Knot.vertices[Knot.numverts].z);

      Knot.colors[Knot.numverts].r := ((2/dist)+(0.5*sin(Beta)+0.4))/2.0;
      Knot.colors[Knot.numverts].g := ((2/dist)+(0.5*sin(Beta)+0.4))/2.0;
      Knot.colors[Knot.numverts].b := ((2/dist)+(0.5*sin(Beta)+0.4))/2.0;

      Inc(Knot.numverts);
    end;
  end;

  for Count1 := 0 to SCALING_FACTOR2-1 do
  begin
    Index1 := Count1 * SCALING_FACTOR1;
    Index2 := Index1 + SCALING_FACTOR1;
    Index2 := Index2 MOD Knot.numverts;
    Rotation := 0;
    MinDistance := (Knot.vertices[Index1].x - Knot.vertices[Index2].x) * (Knot.vertices[Index1].x - Knot.vertices[Index2].x) +
                   (Knot.vertices[Index1].y - Knot.vertices[Index2].y) * (Knot.vertices[Index1].y - Knot.vertices[Index2].y) +
                   (Knot.vertices[Index1].z - Knot.vertices[Index2].z) * (Knot.vertices[Index1].z - Knot.vertices[Index2].z);
    for Count2 := 1 to SCALING_FACTOR1-1 do
    begin
      Index2 := Count2 + Index1 + SCALING_FACTOR1;
      if Count1 = SCALING_FACTOR2 - 1 then
        Index2 := Count2;
      Distance := (Knot.vertices[Index1].x - Knot.vertices[Index2].x) * (Knot.vertices[Index1].x - Knot.vertices[Index2].x) +
                  (Knot.vertices[Index1].y - Knot.vertices[Index2].y) * (Knot.vertices[Index1].y - Knot.vertices[Index2].y) +
                  (Knot.vertices[Index1].z - Knot.vertices[Index2].z) * (Knot.vertices[Index1].z - Knot.vertices[Index2].z);
      if Distance < MinDistance then
      begin
        MinDistance := Distance;
        Rotation := Count2;
      end;
    end;

    for Count2 := 0 to SCALING_FACTOR1 -1 do
    begin
      Knot.faces[4*(Index1+Count2)+0] := Index1 + Count2;

      Index2 := Count2 + 1;
      Index2 := Index2 MOD SCALING_FACTOR1;
      Knot.faces[4*(Index1+Count2)+1] := Index1 + Index2;

      Index2 := Round(Count2 + Rotation + 1);
      Index2 := Index2 MOD SCALING_FACTOR1;
      Knot.faces[4*(Index1+Count2)+2] := (Index1 + Index2 + SCALING_FACTOR1) MOD  Knot.numverts;

      Index2 := Round(Count2 + Rotation);
      Index2 := Index2 MOD SCALING_FACTOR1;

      Knot.faces[4*(Index1+Count2)+3] := (Index1 + Index2 + SCALING_FACTOR1) MOD Knot.numverts;
      Inc(Knot.numfaces);
    end;
  end;

  KnotDL :=glGenLists(1);
  glNewList(KnotDL, GL_COMPILE);
    glBegin(GL_QUADS);
      J := Knot.numfaces*4;
      for I :=0 to J do
      begin
        Count1 := Knot.faces[i];
        glColor3f(Knot.Colors[Count1].r*2, Knot.Colors[Count1].g, Knot.Colors[Count1].b);
        glVertex3f(Knot.vertices[Count1].x, Knot.vertices[Count1].y,  Knot.vertices[Count1].z);
      end;
    glEnd();
  glEndList();
end;


{------------------------------------------------------------------}
{  Function to draw the actual scene                               }
{------------------------------------------------------------------}
procedure glDraw;
begin
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT or GL_STENCIL_BUFFER_BIT);    // Clear The Screen And The Depth Buffer
  glLoadIdentity();                                       // Reset The View
  glTranslatef(0.0,0.0,-20);

  glDisable(GL_TEXTURE_2D);                     // Enable Texture Mapping

  // Create a knot area in the stencil buffer
  glStencilFunc(GL_ALWAYS, 1, 1);
  glStencilOp(GL_REPLACE, GL_REPLACE, GL_REPLACE);

  glPushMatrix();
    glRotatef(ElapsedTime/10, 1, 0, 0);
    glRotatef(ElapsedTime/8, 0, 1, 0);
    glCallList(KnotDL);    // draw the knot area
  glPopMatrix;

  // only draw the square where the knot area is in the stencil buffer
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);    // Clear The Screen And The Depth Buffer
  glStencilFunc(GL_EQUAL, 1, 1);
  glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

  glColor3f(1, 1, 1);
  glEnable(GL_TEXTURE_2D);
  glBegin(GL_QUADS);
    glTexCoord2f(0.0, 0.0); glVertex3f(-5.0, -5.0,  0.0);
    glTexCoord2f(1.0, 0.0); glVertex3f( 5.0, -5.0,  0.0);
    glTexCoord2f(1.0, 1.0); glVertex3f( 5.0,  5.0,  0.0);
    glTexCoord2f(0.0, 1.0); glVertex3f(-5.0,  5.0,  0.0);
  glEnd;
end;


{------------------------------------------------------------------}
{  Initialise OpenGL                                               }
{------------------------------------------------------------------}
procedure glInit;
begin
  glClearColor(0.0, 0.0, 0.0, 0.0);        // Black Background
  glShadeModel(GL_SMOOTH);                 // Enables Smooth Color Shading
  glClearDepth(1.0);                       // Depth Buffer Setup
  glEnable(GL_DEPTH_TEST);                 // Enable Depth Testing
  glDepthFunc(GL_LESS);                    // The Type Of Depth Test To Do
  glEnable(GL_STENCIL_TEST);    {** ENABLE STENCIL BUFFER **}

  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);   //Realy Nice perspective calculations

  createKnot(2, 4.0, 1.0);

  glEnable(GL_TEXTURE_2D);                     // Enable Texture Mapping
  LoadTexture('background.bmp', BackgroundTex);
  glBindTexture(GL_TEXTURE_2D, BackgroundTex);
end;


{------------------------------------------------------------------}
{  Handle window resize                                            }
{------------------------------------------------------------------}
procedure glResizeWnd(Width, Height : Integer);
begin
  if (Height = 0) then                // prevent divide by zero exception
    Height := 1;
  glViewport(0, 0, Width, Height);    // Set the viewport for the OpenGL window
  glMatrixMode(GL_PROJECTION);        // Change Matrix Mode to Projection
  glLoadIdentity();                   // Reset View
  gluPerspective(45.0, Width/Height, 1.0, 100.0);  // Do the perspective calculations. Last value = max clipping depth

  glMatrixMode(GL_MODELVIEW);         // Return to the modelview matrix
  glLoadIdentity();                   // Reset View
end;


{------------------------------------------------------------------}
{  Processes all the keystrokes                                    }
{------------------------------------------------------------------}
procedure ProcessKeys;
begin
end;


{------------------------------------------------------------------}
{  Determines the applications response to the messages received  }
{------------------------------------------------------------------}
function WndProc(hWnd: HWND; Msg: UINT;  wParam: WPARAM;  lParam: LPARAM): LRESULT; stdcall;
var
  Buff: array[0..MAX_PATH] of Char;
begin
  case (Msg) of
    WM_CREATE:
      begin
        // Insert stuff you want executed when the program starts
      end;
    WM_CLOSE:
      begin
        PostQuitMessage(0);
        Result := 0
      end;
    WM_KEYDOWN:       // Set the pressed key (wparam) to equal true so we can check if its pressed
      begin
        keys[wParam] := True;
        Result := 0;
      end;
    WM_KEYUP:         // Set the released key (wparam) to equal false so we can check if its pressed
      begin
        keys[wParam] := False;
        Result := 0;
      end;
    WM_SIZE:          // Resize the window with the new width and height
      begin
        glResizeWnd(LOWORD(lParam),HIWORD(lParam));
        Result := 0;
      end;
    WM_TIMER :                     // Add code here for all timers to be used.
      begin
        if wParam = FPS_TIMER then
        begin
          FPSCount :=Round(FPSCount * 1000/FPS_INTERVAL);   // calculate to get per Second incase intercal is less or greater than 1 second
          SetWindowText(h_Wnd, StrPCopy(Buff, WND_TITLE + '   [' + intToStr(FPSCount) + ' FPS]'));
          FPSCount := 0;
          Result := 0;
        end;
      end;
    else
      Result := DefWindowProc(hWnd, Msg, wParam, lParam);    // Default result if nothing happens
  end;
end;


{---------------------------------------------------------------------}
{  Properly destroys the window created at startup (no memory leaks)  }
{---------------------------------------------------------------------}
procedure glKillWnd(Fullscreen : Boolean);
begin
  if Fullscreen then             // Change back to non fullscreen
  begin
    ChangeDisplaySettings(TDeviceMode(nil^), 0);
    ShowCursor(True);
  end;

  // Makes current rendering context not current, and releases the device
  // context that is used by the rendering context.
  if (not wglMakeCurrent(h_DC, 0)) then
    MessageBox(0, 'Release of DC and RC failed!', 'Error', MB_OK or MB_ICONERROR);

  // Attempts to delete the rendering context
  if (not wglDeleteContext(h_RC)) then
  begin
    MessageBox(0, 'Release of rendering context failed!', 'Error', MB_OK or MB_ICONERROR);
    h_RC := 0;
  end;

  // Attemps to release the device context
  if ((h_DC = 1) and (ReleaseDC(h_Wnd, h_DC) <> 0)) then
  begin
    MessageBox(0, 'Release of device context failed!', 'Error', MB_OK or MB_ICONERROR);
    h_DC := 0;
  end;

  // Attempts to destroy the window
  if ((h_Wnd <> 0) and (not DestroyWindow(h_Wnd))) then
  begin
    MessageBox(0, 'Unable to destroy window!', 'Error', MB_OK or MB_ICONERROR);
    h_Wnd := 0;
  end;

  // Attempts to unregister the window class
  if (not UnRegisterClass('OpenGL', hInstance)) then
  begin
    MessageBox(0, 'Unable to unregister window class!', 'Error', MB_OK or MB_ICONERROR);
    hInstance := 0;
  end;
end;


{--------------------------------------------------------------------}
{  Creates the window and attaches a OpenGL rendering context to it  }
{--------------------------------------------------------------------}
function glCreateWnd(Width, Height : Integer; Fullscreen : Boolean; PixelDepth : Integer) : Boolean;
var
  wndClass : TWndClass;         // Window class
  dwStyle : DWORD;              // Window styles
  dwExStyle : DWORD;            // Extended window styles
  dmScreenSettings : TDeviceMode;   // Screen settings (fullscreen, etc...)
  PixelFormat : GLuint;         // Settings for the OpenGL rendering
  h_Instance : HINST;           // Current instance
  pfd : TPIXELFORMATDESCRIPTOR;  // Settings for the OpenGL window
begin
  h_Instance := GetModuleHandle(nil);       //Grab An Instance For Our Window
  ZeroMemory(@wndClass, SizeOf(wndClass));  // Clear the window class structure

  with wndClass do                    // Set up the window class
  begin
    style         := CS_HREDRAW or    // Redraws entire window if length changes
                     CS_VREDRAW or    // Redraws entire window if height changes
                     CS_OWNDC;        // Unique device context for the window
    lpfnWndProc   := @WndProc;        // Set the window procedure to our func WndProc
    hInstance     := h_Instance;
    hCursor       := LoadCursor(0, IDC_ARROW);
    lpszClassName := 'OpenGL';
  end;
  ShowCursor(FALSE);

  if (RegisterClass(wndClass) = 0) then  // Attemp to register the window class
  begin
    MessageBox(0, 'Failed to register the window class!', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit
  end;

  // Change to fullscreen if so desired
  if Fullscreen then
  begin
    ZeroMemory(@dmScreenSettings, SizeOf(dmScreenSettings));
    with dmScreenSettings do begin              // Set parameters for the screen setting
      dmSize       := SizeOf(dmScreenSettings);
      dmPelsWidth  := Width;                    // Window width
      dmPelsHeight := Height;                   // Window height
      dmBitsPerPel := PixelDepth;               // Window color depth
      dmFields     := DM_PELSWIDTH or DM_PELSHEIGHT or DM_BITSPERPEL;
    end;

    // Try to change screen mode to fullscreen
    if (ChangeDisplaySettings(dmScreenSettings, CDS_FULLSCREEN) = DISP_CHANGE_FAILED) then
    begin
      MessageBox(0, 'Unable to switch to fullscreen!', 'Error', MB_OK or MB_ICONERROR);
      Fullscreen := False;
    end;
  end;

  // If we are still in fullscreen then
  if (Fullscreen) then
  begin
    dwStyle := WS_POPUP or                // Creates a popup window
               WS_CLIPCHILDREN            // Doesn't draw within child windows
               or WS_CLIPSIBLINGS;        // Doesn't draw within sibling windows
    dwExStyle := WS_EX_APPWINDOW;         // Top level window
    ShowCursor(False);                    // Turn of the cursor (gets in the way)
  end
  else
  begin
    dwStyle := WS_OVERLAPPEDWINDOW or     // Creates an overlapping window
               WS_CLIPCHILDREN or         // Doesn't draw within child windows
               WS_CLIPSIBLINGS;           // Doesn't draw within sibling windows
    dwExStyle := WS_EX_APPWINDOW or       // Top level window
                 WS_EX_WINDOWEDGE;        // Border with a raised edge
  end;

  // Attempt to create the actual window
  h_Wnd := CreateWindowEx(dwExStyle,      // Extended window styles
                          'OpenGL',       // Class name
                          WND_TITLE,      // Window title (caption)
                          dwStyle,        // Window styles
                          0, 0,           // Window position
                          Width, Height,  // Size of window
                          0,              // No parent window
                          0,              // No menu
                          h_Instance,     // Instance
                          nil);           // Pass nothing to WM_CREATE
  if h_Wnd = 0 then
  begin
    glKillWnd(Fullscreen);                // Undo all the settings we've changed
    MessageBox(0, 'Unable to create window!', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Try to get a device context
  h_DC := GetDC(h_Wnd);
  if (h_DC = 0) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to get a device context!', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Settings for the OpenGL window
  with pfd do
  begin
    nSize           := SizeOf(TPIXELFORMATDESCRIPTOR); // Size Of This Pixel Format Descriptor
    nVersion        := 1;                    // The version of this data structure
    dwFlags         := PFD_DRAW_TO_WINDOW    // Buffer supports drawing to window
                       or PFD_SUPPORT_OPENGL // Buffer supports OpenGL drawing
                       or PFD_DOUBLEBUFFER;  // Supports double buffering
    iPixelType      := PFD_TYPE_RGBA;        // RGBA color format
    cColorBits      := PixelDepth;           // OpenGL color depth
    cRedBits        := 0;                    // Number of red bitplanes
    cRedShift       := 0;                    // Shift count for red bitplanes
    cGreenBits      := 0;                    // Number of green bitplanes
    cGreenShift     := 0;                    // Shift count for green bitplanes
    cBlueBits       := 0;                    // Number of blue bitplanes
    cBlueShift      := 0;                    // Shift count for blue bitplanes
    cAlphaBits      := 0;                    // Not supported
    cAlphaShift     := 0;                    // Not supported
    cAccumBits      := 0;                    // No accumulation buffer
    cAccumRedBits   := 0;                    // Number of red bits in a-buffer
    cAccumGreenBits := 0;                    // Number of green bits in a-buffer
    cAccumBlueBits  := 0;                    // Number of blue bits in a-buffer
    cAccumAlphaBits := 0;                    // Number of alpha bits in a-buffer
    cDepthBits      := 16;                   // Specifies the depth of the depth buffer
    cStencilBits    := 1;                    // Turn off stencil buffer
    cAuxBuffers     := 0;                    // Not supported
    iLayerType      := PFD_MAIN_PLANE;       // Ignored
    bReserved       := 0;                    // Number of overlay and underlay planes
    dwLayerMask     := 0;                    // Ignored
    dwVisibleMask   := 0;                    // Transparent color of underlay plane
    dwDamageMask    := 0;                     // Ignored
  end;

  // Attempts to find the pixel format supported by a device context that is the best match to a given pixel format specification.
  PixelFormat := ChoosePixelFormat(h_DC, @pfd);
  if (PixelFormat = 0) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to find a suitable pixel format', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Sets the specified device context's pixel format to the format specified by the PixelFormat.
  if (not SetPixelFormat(h_DC, PixelFormat, @pfd)) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to set the pixel format', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Create a OpenGL rendering context
  h_RC := wglCreateContext(h_DC);
  if (h_RC = 0) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to create an OpenGL rendering context', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Makes the specified OpenGL rendering context the calling thread's current rendering context
  if (not wglMakeCurrent(h_DC, h_RC)) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to activate OpenGL rendering context', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Initializes the timer used to calculate the FPS
  SetTimer(h_Wnd, FPS_TIMER, FPS_INTERVAL, nil);

  // Settings to ensure that the window is the topmost window
  ShowWindow(h_Wnd, SW_SHOW);
  SetForegroundWindow(h_Wnd);
  SetFocus(h_Wnd);

  // Ensure the OpenGL window is resized properly
  glResizeWnd(Width, Height);
  glInit();

  Result := True;
end;


{--------------------------------------------------------------------}
{  Main message loop for the application                             }
{--------------------------------------------------------------------}
function WinMain(hInstance : HINST; hPrevInstance : HINST;
                 lpCmdLine : PChar; nCmdShow : Integer) : Integer; stdcall;
var
  msg : TMsg;
  finished : Boolean;
  DemoStart, LastTime : DWord;
begin
  finished := False;

  // Perform application initialization:
  if not glCreateWnd(800, 600, FALSE, 32) then
  begin
    Result := 0;
    Exit;
  end;

  DemoStart := GetTickCount();            // Get Time when demo started

  // Main message loop:
  while not finished do
  begin
    if (PeekMessage(msg, 0, 0, 0, PM_REMOVE)) then // Check if there is a message for this window
    begin
      if (msg.message = WM_QUIT) then     // If WM_QUIT message received then we are done
        finished := True
      else
      begin                               // Else translate and dispatch the message to this window
        TranslateMessage(msg);
        DispatchMessage(msg);
      end;
    end
    else
    begin
      Inc(FPSCount);                      // Increment FPS Counter

      LastTime :=ElapsedTime;
      ElapsedTime :=GetTickCount() - DemoStart;     // Calculate Elapsed Time
      ElapsedTime :=(LastTime + ElapsedTime) DIV 2; // Average it out for smoother movement

      glDraw();                           // Draw the scene
      SwapBuffers(h_DC);                  // Display the scene

      if (keys[VK_ESCAPE]) then           // If user pressed ESC then set finised TRUE
        finished := True
      else
        ProcessKeys;                      // Check for any other key Pressed
    end;
  end;
  glKillWnd(FALSE);
  Result := msg.wParam;
end;


begin
  WinMain( hInstance, hPrevInst, CmdLine, CmdShow );
end.
